# ConfigData.pm - Will pull settings from an XML config file into a format usable by the system.
# Created by James A. Pattie.  Copyright (c) 2001-2004, Xperience, Inc.

package PCXFireWall::ConfigData;

use strict;
use XML::LibXML;
use vars qw ($AUTOLOAD @ISA @EXPORT $VERSION);

require Exporter;
@ISA = qw(Exporter AutoLoader);
@EXPORT = qw();

$VERSION = "2.5";

# new
sub new
{
  my $that = shift;
  my $class = ref($that) || $that;
  my $self = bless {}, $class;
  my $errStr = "ConfigData->new()  - Error:";

  $self->{data} = {};
  $self->{errorCodes} = {
    0  => "_info_ not defined",
    1  => "_info_->module not defined",
    2  => "_info_->version not defined",
    3  => "_info_->confFile not defined",
    4  => "module '%s' unknown",
    5  => "config not defined",
    6  => "config->networkCommand not defined",
    7  => "config->networkCommand = '%s' is invalid",
    8  => "config->mangleRules not defined",
    9  => "config->mangleRules = '%s' is invalid",
    10 => "config->zones not defined",
    11 => "config->zones->%s not defined",
    12 => "config->zones->%s = '%s' is invalid",
    13 => "config->logInfo not defined",
    14 => "config->logInfo->%s not defined",
    15 => "config->logInfo->%s = '%s' is invalid",
    16 => "config->limit not defined",
    17 => "config->limitRates not defined",
    18 => "config->limit->log->%s not defined",
    19 => "config->limitRates->log->%s not defined",
    20 => "config->limitRates->log->%s = '%s' is invalid",
    21 => "config->limit->icmp not defined",
    22 => "config->limitRates->icmp not defined",
    23 => "config->limitRates->icmp = '%s' is invalid",
    24 => "config->networks not defined",
    25 => "No networks configured",
    26 => "config->reserved not defined",
    27 => "No reserved entries configured",
    28 => "config->interfaces not defined",
    29 => "config->interfaces->%s not defined",
    30 => "No interfaces defined for zone = '%s'",
    31 => "services not defined",
    32 => "services->servicesOrder array doesn't exist!  Internal error!",
    33 => "No services defined",
    34 => "No paths defined for service = '%s'",
    35 => "services->%s->zonePathOrder array does not exist!  Internal error!",
    36 => "No nodes defined for path = '%s', service = '%s'",
    37 => "services->%s->%s->infos array does not exist!  Internal error!",
    38 => "No info entries defined for path = '%s', service = '%s'",
    39 => "No aliases defined",
    40 => "config->ecn not defined",
    41 => "config->ecn = '%s' is invalid",
    42 => "config->moduleSupport not defined",
    43 => "config->moduleSupport->modular not defined",
    44 => "config->moduleSupport->modular = '%s' is invalid",
    45 => "config->moduleSupport->%s not defined",
    46 => "config->moduleSupport->%s->%s not defined",
    47 => "config->moduleSupport->%s->%s->%s not defined",
    48 => "config->moduleSupport->%s->%s->%s = '%s' is invalid",
    49 => "config->validityChecks not defined",
    50 => "config->validityChecks->%s not defined",
    51 => "config->validityChecks->%s = '%s' is invalid",
    52 => "version = '%s' is invalid for module = '%s'",
    53 => "No serviceInfo's defined for service = '%s'",
    54 => "services->%s->infos array does not exist!  Internal error!",
    55 => "serviceInfo->%s must exist for service = '%s'",
    56 => "paths not defined",
    57 => "paths->servicesOrder array doesn't exist!  Internal error!",
    58 => "paths->%s->%s->%s doesn't exist!",
    59 => "'%s' is an invalid networks index!",
    60 => "networks->%s->%s doesn't exist!",
    61 => "reserved[%s]->%s does not exist!",
    62 => "interfaces->%s->%s->%s does not exist!",
    63 => "interfaces->%s->%s->%s = '%s' is invalid!",
    64 => "interfaces->%s->%s->%s must be specified!",
    65 => "reserved[%s]->%s = '%s' is invalid!",
    66 => "networks->%s->%s = '%s' is invalid!",
    67 => "paths->%s->%s->%s = '%s' is invalid!",
    68 => "All %s entries are disabled!  You need at least one enabled.",
    69 => "reserved[%s]->%s = '%s' is duplicated!",
    70 => "Interface '%s' is in inactive zone '%s' and can not be boundTo for ipsec interface '%s'!",
    71 => "Interface '%s' in zone '%s' does not have ipsec enabled!  ipsec interface '%s'",
    72 => "boundTo interface '%s' could not be found for ipsec interface '%s'!",
    73 => "Interface '%s' in zone '%s' disabled!  ipsec interface '%s'",
    74 => "Network type '%s' is invalid for interface '%s' in zone '%s'!",
    75 => "All %s defined for interface '%s' are disabled!",
    76 => "Network type '%s' is invalid for ipsecLimitFrom '%s', interface '%s' in zone '%s'!",
    77 => "Network '%s' does not exist for %s '%s', %s '%s', path <%s> in service '%s'!",
    78 => "Network type '%s' is invalid for %s '%s', %s '%s', path <%s> in service '%s'!",
    79 => "All %s entries are disabled for %s '%s', %s '%s', path <%s> in service '%s'!",
    80 => "DNAT destination '%s' is disabled for patch <%s> in service '%s'!",
    81 => "config->%s not defined",
    82 => "config->%s = '%s' is invalid",
    83 => "zone '%s', alias '%s', interface '%s' is duplicated in zone '%s', alias '%s'!",
  };

  return $self;
}

sub AUTOLOAD
{
  my $self = shift;
  my $type = ref($self) || die "$self is not an object";
  my $name = $AUTOLOAD;
  $name =~ s/.*://; # strip fully-qualified portion
  unless (exists $self->{$name})
  {
    die "Can't access `$name' field in object of class $type";
  }
  if (@_)
  {
    return $self->{$name} = shift;
  }
  else
  {
    return $self->{$name};
  }
}

sub isValid
{
  my $self = shift;
  my @errors = ();
  my %data = %{$self->{data}};
  my $module = "";
  my $version = "";

  my $errStr = "This is being removed: ";
  my $strict = 0;

  if (not exists $data{'_info_'})
  {
    push @errors, $self->{errorCodes}->{0};
  }
  if (not exists $data{'_info_'}{module})
  {
    push @errors, $self->{errorCodes}->{1};
  }
  else
  {
    $module = $data{'_info_'}{module};
  }
  if (not exists $data{'_info_'}{version})
  {
    push @errors, $self->{errorCodes}->{2};
  }
  else
  {
    $version = $data{'_info_'}{version};
  }
  if (not exists $data{'_info_'}{confFile})
  {
    push @errors, $self->{errorCodes}->{3};
  }
  if ($module =~ /^(VeryTight2?)$/)
  {
    if ($version =~ /^(2\.2)$/)
    {
      # validate version 2.x data structure.
      if (not exists $data{config})
      {
        push @errors, $self->{errorCodes}->{5};
      }
      if (not exists $data{config}{networkCommand})
      {
        push @errors, $self->{errorCodes}->{6};
      }
      elsif ($data{config}{networkCommand} !~ /^(ifconfig|ip)$/)
      {
        push @errors, sprintf($self->{errorCodes}->{7}, $data{config}{networkCommand});
      }
      if (not exists $data{config}{mangleRules})
      {
        push @errors, $self->{errorCodes}->{8};
      }
      elsif ($data{config}{mangleRules} !~ /^(0|1)$/)
      {
        push @errors, sprintf($self->{errorCodes}->{9}, $data{config}{mangleRules});
      }
      if (not exists $data{config}{ecn})
      {
        push @errors, $self->{errorCodes}->{40};
      }
      elsif ($data{config}{ecn} !~ /^(0|1)$/)
      {
        push @errors, sprintf($self->{errorCodes}->{41}, $data{config}{ecn});
      }
      if (not exists $data{config}{bridgeSupport})
      {
        push @errors, sprintf($self->{errorCodes}->{81}, "bridgeSupport");
      }
      elsif ($data{config}{bridgeSupport} !~ /^(0|1)$/)
      {
        push @errors, sprintf($self->{errorCodes}->{82}, "bridgeSupport", $data{config}{bridgeSupport});
      }
      if (not exists $data{config}{snortInlineSupport})
      {
        push @errors, sprintf($self->{errorCodes}->{81}, "snortInlineSupport");
      }
      elsif ($data{config}{snortInlineSupport} !~ /^(0|1)$/)
      {
        push @errors, sprintf($self->{errorCodes}->{82}, "snortInlineSupport", $data{config}{snortInlineSupport});
      }
      if (not exists $data{config}{dynamicInterfacesMode})
      {
        push @errors, sprintf($self->{errorCodes}->{81}, "dynamicInterfacesMode");
      }
      elsif ($data{config}{dynamicInterfacesMode} !~ /^(lookupIP|ignoreIP)$/)
      {
        push @errors, sprintf($self->{errorCodes}->{82}, "dynamicInterfacesMode", $data{config}{dynamicInterfacesMode});
      }
      if (not exists $data{config}{zones})
      {
        push @errors, $self->{errorCodes}->{10};
      }
      else
      {
        # validate zones entries.
        foreach my $entry ("internal", "ipsec", "dmz", "dialin")
        {
          if (not exists $data{config}{zones}{$entry})
          {
            push @errors, sprintf($self->{errorCodes}->{11}, $entry);
          }
          elsif ($data{config}{zones}{$entry} !~ /^(0|1)$/)
          {
            push @errors, sprintf($self->{errorCodes}->{12}, $entry, $data{config}{zones}{$entry});
          }
        }
      }
      if (not exists $data{config}{logInfo})
      {
        push @errors, $self->{errorCodes}->{13};
      }
      else
      {
        foreach my $entry ("prefix", "level")
        {
          if (not exists $data{config}{logInfo}{$entry})
          {
            push @errors, $self->{errorCodes}->{14};
          }
          else
          {
            if ($entry eq "level")
            {
              if ($data{config}{logInfo}{$entry} !~ /^(debug|info|notice|warning|err|crit|alert|emerg)$/)
              {
                push @errors, sprintf($self->{errorCodes}->{15}, $entry, $data{config}{logInfo}{$entry});
              }
            }
            else
            {
              if (length $data{config}{logInfo}{$entry} == 0)
              {
                push @errors, sprintf($self->{errorCodes}->{15}, $entry, $data{config}{logInfo}{$entry});
              }
            }
          }
        }
      }
      if (not exists $data{config}{limit})
      {
        push @errors, $self->{errorCodes}->{16};
      }
      else
      {
        if (not exists $data{config}{limitRates})
        {
          push @errors, $self->{errorCodes}->{17};
        }
        else
        {
          if (not exists $data{config}{limitRates}{icmp})
          {
            push @errors, $self->{errorCodes}->{22};
          }
          elsif ($data{config}{limitRates}{icmp} !~ /^(0|\d+\/([mshd]|minute|second|hour|day))$/)
          {
            push @errors, sprintf($self->{errorCodes}->{23}, $data{config}{limitRates}{icmp});
          }
          foreach my $entry ("taint", "reserved", "policy", "reject", "normal")
          {
            if (not exists $data{config}{limit}{log}{$entry})
            {
              push @errors, sprintf($self->{errorCodes}->{18}, $entry);
            }
            if (not exists $data{config}{limitRates}{log}{$entry})
            {
              push @errors, sprintf($self->{errorCodes}->{19}, $entry);
            }
            elsif ($data{config}{limitRates}{log}{$entry} !~ /^(0|\d+\/([mshd]|minute|second|hour|day))$/)
            {
              push @errors, sprintf($self->{errorCodes}->{20}, $entry, $data{config}{limitRates}{log}{$entry});
            }
          }
          if (not exists $data{config}{limit}{icmp})
          {
            push @errors, $self->{errorCodes}->{21};
          }
        }
      }
      if (not exists $data{config}{networks})
      {
        push @errors, $self->{errorCodes}->{24};
      }
      elsif (scalar keys(%{$data{config}{networks}}) == 0)
      {
        push @errors, $self->{errorCodes}->{25};
      }
      else
      {
        # validate the specified networks
        my $disabled = 0;
        foreach my $index (keys %{$data{config}{networks}})
        {
          if ($index !~ /^(\d+)$/)
          {
            push @errors, sprintf($self->{errorCodes}->{59}, $index);
          }
          foreach my $name ("host", "comment", "active", "type")
          {
            if (not exists $data{config}{networks}{$index}{$name})
            {
              push @errors, sprintf($self->{errorCodes}->{60}, $index, $name);
            }
          }
          if (length $data{config}{networks}{$index}{host} == 0)
          {
            push @errors, sprintf($self->{errorCodes}->{66}, $index, "host", $data{config}{networks}{$index}{host});
          }
          if ($data{config}{networks}{$index}{active} !~ /^(0|1)$/)
          {
            push @errors, sprintf($self->{errorCodes}->{66}, $index, "active", $data{config}{networks}{$index}{active});
          }
          if ($data{config}{networks}{$index}{type} !~ /^(normal|mac)$/)
          {
            push @errors, sprintf($self->{errorCodes}->{66}, $index, "type", $data{config}{networks}{$index}{type});
          }
          $disabled++ if (!$data{config}{networks}{$index}{active});
        }
        if ($disabled == scalar keys %{$data{config}{networks}})
        {
          push @errors, sprintf($self->{errorCodes}->{68}, "networks");
        }
      }
      if (not exists $data{config}{reserved})
      {
        push @errors, $self->{errorCodes}->{26};
      }
      elsif (scalar @{$data{config}{reserved}} == 0)
      {
        push @errors, $self->{errorCodes}->{27};
      }
      else
      {
        # validate the reserved networks specified.
        my $counter = 0;
        foreach my $node (@{$data{config}{reserved}})
        {
          foreach my $name ("host", "comment", "active", "limitTo")
          {
            if (not exists $node->{$name})
            {
              push @errors, sprintf($self->{errorCodes}->{61}, $counter, $name);
            }
          }
          if (length $node->{host} == 0)
          {
            push @errors, sprintf($self->{errorCodes}->{65}, $counter, "host", $node->{host});
          }
          if ($node->{active} !~ /^(0|1)$/)
          {
            push @errors, sprintf($self->{errorCodes}->{65}, $counter, "active", $node->{active});
          }
          my %aliases = ();
          foreach my $alias (@{$node->{limitTo}})
          {
            if (not exists $data{config}{interfaces}{external}{$alias})
            {
              push @errors, sprintf($self->{errorCodes}->{65}, $counter, "limitTo", $alias);
            }
            if (exists $aliases{$alias})
            {
              push @errors, sprintf($self->{errorCodes}->{69}, $counter, "limitTo", $alias);
            }
            else
            {
              $aliases{$alias} = 1;
            }
          }
          $counter++;
        }
      }
      if (not exists $data{config}{interfaces})
      {
        push @errors, $self->{errorCodes}->{28};
      }
      else
      {
        if (not exists $data{config}{interfaces}{aliasLookup})
        {
          push @errors, sprintf($self->{errorCodes}->{29}, "aliasLookup");
        }
        elsif (scalar keys %{$data{config}{interfaces}{aliasLookup}} == 0)
        {
          push @errors, $self->{errorCodes}->{39};
        }
        foreach my $zone ("external", "internal", "ipsec", "dmz", "dialin")
        {
          if (not exists $data{config}{interfaces}{$zone})
          {
            push @errors, sprintf($self->{errorCodes}->{29}, $zone);
          }
          elsif (scalar keys %{$data{config}{interfaces}{$zone}} == 0)
          {
            next if ($zone !~ /^(external)$/ && $data{config}{zones}{$zone} == 0);  # not being used.
            push @errors, sprintf($self->{errorCodes}->{30}, $zone);
          }
          else
          {
            # validate the interfaces in this zone.
            my $disabled = 0;
            foreach my $alias (keys %{$data{config}{interfaces}{$zone}})
            {
              my $info = $data{config}{interfaces}{$zone}{$alias};
              foreach my $name ("name", "network", "comment", "active")
              {
                if (not exists $info->{$name})
                {
                  push @errors, sprintf($self->{errorCodes}->{62}, $zone, $alias, $name);
                }
              }
              if ($info->{name} eq "")
              {
                push @errors, sprintf($self->{errorCodes}->{63}, $zone, $alias, "name", $info->{name});
              }
              # make sure this is not duplicated anywhere.
              foreach my $tmpZone ("internal", "external", "ipsec", "dmz", "dialin")
              {
                foreach my $tmpAlias (keys %{$data{config}{interfaces}{$tmpZone}})
                {
                  if ($data{config}{interfaces}{$tmpZone}{$tmpAlias}{name} eq $info->{name} && ($alias ne $tmpAlias))
                  {
                    push @errors, sprintf($self->{errorCodes}->{83}, $zone, $alias, $info->{name}, $tmpZone, $tmpAlias);
                  }
                }
              }
              if (scalar @{$info->{network}} == 0)
              {
                push @errors, sprintf($self->{errorCodes}->{64}, $zone, $alias, "network");
              }
              else
              {
                my $noNetworks = 0;
                foreach my $index (@{$info->{network}})
                {
                  if (!exists $data{config}{networks}{$index})
                  {
                    push @errors, sprintf($self->{errorCodes}->{63}, $zone, $alias, "network", $index);
                  }
                  elsif ($data{config}{networks}{$index}{type} ne "normal")
                  {
                    push @errors, sprintf($self->{errorCodes}->{74}, $data{config}{networks}{$index}{type}, $alias, $zone);
                  }
                  $noNetworks++ if (!$data{config}{networks}{$index}{active});
                }
                if ($noNetworks == scalar @{$info->{network}} && $info->{active} && ($zone !~ /^(external)$/ && $data{config}{zones}{$zone}))
                {
                  push @errors, sprintf($self->{errorCodes}->{75}, "networks", $alias);
                }
              }
              if ($info->{active} !~ /^(1|0)$/)
              {
                push @errors, sprintf($self->{errorCodes}->{63}, $zone, $alias, "active", $info->{active});
              }
              $disabled++ if (!$info->{active});
              if ($info->{proxyarp} !~ /^(0|1)$/)
              {
                push @errors, sprintf($self->{errorCodes}->{63}, $zone, $alias, "proxyarp", $info->{proxyarp});
              }
              if ($zone =~ /^(external|internal|dmz|dialin)$/)
              {
                foreach my $name ("ip", "ipsec", "ipsecLimitFrom")
                {
                  if (not exists $info->{$name})
                  {
                    push @errors, sprintf($self->{errorCodes}->{62}, $zone, $alias, $name);
                  }
                }
                if ($info->{ip} !~ /^(\d+\.\d+\.\d+\.\d+|dynamic)$/) # very simplified IP check
                {
                  push @errors, sprintf($self->{errorCodes}->{63}, $zone, $alias, "ip", $info->{ip});
                }
                if ($info->{ipsec} !~ /^(0|1)$/)
                {
                  push @errors, sprintf($self->{errorCodes}->{63}, $zone, $alias, "ipsec", $info->{ipsec});
                }
                my $noLimitFrom = 0;
                foreach my $index (@{$info->{ipsecLimitFrom}})
                {
                  if (!exists $data{config}{networks}{$index})
                  {
                    push @errors, sprintf($self->{errorCodes}->{63}, $zone, $alias, "ipsecLimitFrom", $index);
                  }
                  elsif ($data{config}{networks}{$index}{type} !~ /^(normal)$/)
                  {
                    push @errors, sprintf($self->{errorCodes}->{76}, $data{config}{networks}{$index}{type}, $index, $alias, $zone);
                  }
                  $noLimitFrom++ if (!$data{config}{networks}{$index}{active});
                }
                if ($noLimitFrom > 0 && $noLimitFrom == scalar @{$info->{ipsecLimitFrom}} && $info->{active} && ($zone !~ /^(external)$/ && $data{config}{zones}{$zone}))
                {
                  push @errors, sprintf($self->{errorCodes}->{75}, "ipsecLimitFroms", $alias);
                }
              }
              else
              {
                if (!exists $info->{boundTo})
                {
                  push @errors, sprintf($self->{errorCodes}->{62}, $zone, $alias, "boundTo");
                }
                else
                {
                  my $found = 0;
                  foreach my $zone2 ("internal", "external", "dmz", "dialin")
                  {
                    if (exists $data{config}{interfaces}{$zone2}{$info->{boundTo}})
                    {
                      $found = 1;
                      # if the zone is internal, dmz or dialin, make sure it is active!
                      if ($zone2 =~ /^(internal|dmz|dialin)$/)
                      {
                        if (!$data{config}{zones}{$zone2})
                        {
                          push @errors, sprintf($self->{errorCodes}->{70}, $info->{boundTo}, $zone2, $alias);
                        }
                      }
                      # make sure it has ipsec enabled.
                      if ($data{config}{zones}{ipsec} && $data{config}{interfaces}{$zone2}{$info->{boundTo}}{ipsec} != 1)
                      {
                        push @errors, sprintf($self->{errorCodes}->{71}, $info->{boundTo}, $zone2, $alias);
                      }
                      # make sure it is enabled.
                      if ($data{config}{zones}{ipsec} && !$data{config}{interfaces}{$zone2}{$info->{boundTo}}{active} && $info->{active})
                      {
                        push @errors, sprintf($self->{errorCodes}->{73}, $info->{boundTo}, $zone2, $alias);
                      }
                      last;
                    }
                  }
                  if (!$found)
                  {
                    push @errors, sprintf($self->{errorCodes}->{72}, $info->{boundTo}, $alias);
                  }
                }
              }
            }
            if ($disabled == scalar keys %{$data{config}{interfaces}{$zone}} && $data{config}{zones}{$zone})
            {
              push @errors, sprintf($self->{errorCodes}->{68}, "$zone interfaces");
            }
          }
        }
      }

      # verify the moduleSupport data.
      if (not exists $data{config}{moduleSupport})
      {
        push @errors, $self->{errorCodes}->{42};
      }
      else
      {
        if (not exists $data{config}{moduleSupport}{modular})
        {
          push @errors, $self->{errorCodes}->{43};
        }
        elsif ($data{config}{moduleSupport}{modular} !~ /^(0|1)$/)
        {
          push @errors, sprintf($self->{errorCodes}->{44}, $data{config}{moduleSupport}{modular});
        }
        
        # make sure we have the array of user defined modules, though it should be empty by default.
        if (not exists $data{config}{moduleSupport}{modules})
        {
          push @errors, sprintf($self->{errorCodes}->{45}, "modules");
        }
        
        # make sure we have the conntrackers hash, which used to be the modules hash.
        if (not exists $data{config}{moduleSupport}{conntrackers})
        {
          push @errors, sprintf($self->{errorCodes}->{45}, "conntrackers");
        }
        else
        {
          # validate the conntrackers specified, no longer explicitly ftp, irc.
          foreach my $module (keys %{$data{config}{moduleSupport}{conntrackers}})
          {
            if (not exists $data{config}{moduleSupport}{conntrackers}{$module}{active})
            {
              push @errors, sprintf($self->{errorCodes}->{47}, "conntrackers", $module, "active");
            }
            elsif ($data{config}{moduleSupport}{conntrackers}{$module}{active} !~ /^(0|1)$/)
            {
              push @errors, sprintf($self->{errorCodes}->{48}, "conntrackers", $module, "active", $data{config}{moduleSupport}{conntrackers}{$module}{active});
            }
            if (not exists $data{config}{moduleSupport}{conntrackers}{$module}{params})
            {
              push @errors, sprintf($self->{errorCodes}->{47}, "conntrackers", $module, "params");
            }
          }
        }
      }

      # verify the validityChecks section.
      if (not exists $data{config}{validityChecks})
      {
        push @errors, $self->{errorCodes}->{49};
      }
      else
      {
        foreach my $check ("unclean", "tcpFlags")
        {
          if (not exists $data{config}{validityChecks}{$check})
          {
            push @errors, sprintf($self->{errorCodes}->{50}, $check);
          }
          else
          {
            if ($data{config}{validityChecks}{$check} !~ /^(0|1)$/)
            {
              push @errors, sprintf($self->{errorCodes}->{51}, $check, $data{config}{validityChecks}{$check});
            }
          }
        }
      }

      # verify we have a valid services section.
      if (not exists $data{services})
      {
        push @errors, $self->{errorCodes}->{31};
      }
      else
      {
        if (not exists $data{services}{servicesOrder})
        {
          push @errors, $self->{errorCodes}->{32};
        }
        else
        {
          if (scalar keys %{$data{services}} == 1)
          {
            push @errors, $self->{errorCodes}->{33};
          }
          foreach my $service (@{$data{services}{servicesOrder}})
          {
            if (scalar keys %{$data{services}{$service}} < 1)
            {
              push @errors, sprintf($self->{errorCodes}->{53}, $service);
            }
            if (!exists $data{services}{$service}{infos})
            {
              push @errors, ($self->{errorCodes}->{54}, $service);
            }
            foreach my $node (@{$data{services}{$service}{infos}})
            {
              # validate the serviceInfo entry.
              foreach my $name ("proto", "dport", "sport", "icmpType")
              {
                if (not exists $node->{$name})
                {
                  push @errors, sprintf($self->{errorCodes}->{55}, $name, $service);
                }
              }
            }
          }
        }
      }

      # verify the paths info
      if (not exists $data{paths})
      {
        push @errors, $self->{errorCodes}->{56};
      }
      else
      {
        if ($version =~ /^2\.2$/)
        {
          if (not exists $data{paths}{serviceGroupsOrder})
          {
            push @errors, $self->{errorCodes}->{57};
          }
          else
          {
            if (scalar keys %{$data{paths}} == 1)
            {
              push @errors, $self->{errorCodes}->{33};
            }
            foreach my $service (@{$data{paths}{serviceGroupsOrder}})
            {
              if (scalar keys %{$data{paths}{$service}} <= 1)
              {
                push @errors, sprintf($self->{errorCodes}->{34}, $service);
              }
              if (!exists $data{paths}{$service}{zonePathOrder})
              {
                push @errors, ($self->{errorCodes}->{35}, $service);
              }
              foreach my $path (@{$data{paths}{$service}{zonePathOrder}})
              {
                if (scalar @{$data{paths}{$service}{$path}} == 0)
                {
                  push @errors, sprintf($self->{errorCodes}->{36}, $path, $service);
                }
                foreach my $node (@{$data{paths}{$service}{$path}})
                {
                  # make sure we have a service entry first.
                  if (!exists $node->{service})
                  {
                    push @errors, sprintf($self->{errorCodes}->{58}, $service, $path, 'service');
                  }
                  # validate we have at least 1 entry.
                  elsif (scalar @{$node->{service}} == 0)
                  {
                    push @errors, sprintf($self->{errorCodes}->{58}, $service, $path, 'service');
                  }
                  # make sure all the service entries specified exist.
                  else
                  {
                    foreach my $serviceEntry (@{$node->{service}})
                    {
                      if (!exists $data{services}{$serviceEntry})
                      {
                        push @errors, sprintf($self->{errorCodes}->{67}, $service, $path, 'service', $serviceEntry);
                      }
                    }
                  }
                  # we should validate the rest of the path info.
                  if (exists $node->{limitFrom})
                  {
                    my $noLimit = 0;
                    foreach my $index (@{$node->{limitFrom}})
                    {
                      if (!exists $data{config}{networks}{$index})
                      {
                        push @errors, sprintf($self->{errorCodes}->{77}, $index, "i", $node->{i}, "limitFrom", join(",", @{$node->{limitFrom}}), $path, $service);
                      }
                      if ($data{config}{networks}{$index}{type} !~ /^(normal|mac)$/)
                      {
                        push @errors, sprintf($self->{errorCodes}->{78}, $data{config}{networks}{$index}{type}, "i", $node->{i}, "limitFrom", join(",", @{$node->{limitFrom}}), $path, $service);
                      }
                      $noLimit++ if (!$data{config}{networks}{$index}{active});
                    }
                    if ($noLimit > 0 && $noLimit == scalar @{$node->{limitFrom}} && $node->{active})
                    {
                      push @errors, sprintf($self->{errorCodes}->{79}, "limitFrom", "i", $node->{i}, "limitFrom", join(",", @{$node->{limitFrom}}), $path, $service);
                    }
                  }
                  if (exists $node->{limitTo})
                  {
                    my $noLimit = 0;
                    foreach my $index (@{$node->{limitTo}})
                    {
                      if (!exists $data{config}{networks}{$index})
                      {
                        push @errors, sprintf($self->{errorCodes}->{77}, $index, "o", $node->{o}, "limitTo", join(",", @{$node->{limitTo}}), $path, $service);
                      }
                      if ($data{config}{networks}{$index}{type} !~ /^(normal)$/)
                      {
                        push @errors, sprintf($self->{errorCodes}->{78}, $data{config}{networks}{$index}{type}, "o", $node->{o}, "limitTo", join(",", @{$node->{limitTo}}), $path, $service);
                      }
                      $noLimit++ if (!$data{config}{networks}{$index}{active});
                    }
                    if ($noLimit > 0 && $noLimit == scalar @{$node->{limitTo}} && $node->{active})
                    {
                      push @errors, sprintf($self->{errorCodes}->{79}, "limitTo", "o", $node->{o}, "limitTo", join(",", @{$node->{limitTo}}), $path, $service);
                    }
                  }
                  if ($node->{dnat})
                  {
                    if (!$data{config}{networks}{$node->{dnatDestination}}->{active})
                    {
                      push @errors, sprintf($self->{errorCodes}->{80}, $node->{dnatDestination}, $path, $service);
                    }
                  }
                  # I should really validate the i and o attributes, but... :)
                }
              }
            }
          }
        }
      }
    }
    else
    {
      push @errors, sprintf($self->{errorCodes}->{52}, $version, $module);  # version unknown.
    }
  }
  else
  {
    push @errors, sprintf($self->{errorCodes}->{4}, $module); # module unknown.
  }

  return ((scalar @errors > 0 ? 0 : 1), \@errors);
}

sub generateXML
{
  my $self = shift;
  my $result = "";
  my $errStr = "ConfigData->generateXML()  - Error:";
  my %data = %{$self->{data}};

  my @valid = $self->isValid();
  if ($valid[0])
  {
    if ($data{'_info_'}{module} =~ /^(VeryTight2?)$/)
    {
      if ($data{'_info_'}{version} =~ /^(2\.2)$/)
      {
        $result .= <<"END_OF_XML";
<?xml version="1.0" encoding="ISO-8859-1"?>
<firewall module="$data{'_info_'}{module}" moduleVersion="$data{'_info_'}{version}">
END_OF_XML

        # output the <config> section
        $result .= "  <config>\n";
        $result .= "    <networkCommand command=\"$data{config}{networkCommand}\"\/>\n";
        $result .= "    <mangleRules enabled=\"" . ($data{config}{mangleRules} ? "true" : "false") . "\"\/>\n";
        $result .= "    <ecn enabled=\"" . ($data{config}{ecn} ? "true" : "false") . "\"\/>\n";
        $result .= "    <bridgeSupport enabled=\"" . ($data{config}{bridgeSupport} ? "true" : "false") . "\"\/>\n";
        $result .= "    <snortInlineSupport enabled=\"" . ($data{config}{snortInlineSupport} ? "true" : "false") . "\"\/>\n";
        $result .= "    <dynamicInterfaces mode=\"$data{config}{dynamicInterfacesMode}\"\/>  <!-- possible modes are 'lookupIP', 'ignoreIP'.  Use ignoreIP if you are on a laptop and change networks frequently. -->\n";
        $result .= "    <validityChecks";
        foreach my $entry (sort keys %{$data{config}{validityChecks}})
        {
          $result .= " $entry=\"" . ($data{config}{validityChecks}{$entry} ? "true" : "false") . "\"";
        }
        $result .= "\/>\n";
        $result .= "    <moduleSupport kernel=\"" . ($data{config}{moduleSupport}{modular} ? "modular" : "monolithic") . "\">\n";
        foreach my $entry (sort keys %{$data{config}{moduleSupport}{conntrackers}})
        {
          $result .= "      <conntracker name=\"$entry\" active=\"" . ($data{config}{moduleSupport}{conntrackers}{$entry}{active} ? "true" : "false") . "\" ";
          $result .= "params=\"" . $data{config}{moduleSupport}{conntrackers}{$entry}{params} . "\"\/>\n";
        }
        foreach my $entry (@{$data{config}{moduleSupport}{modules}})
        {
          $result .= "      <module name=\"$entry\" \/>\n";
        }
        $result .= "    <\/moduleSupport>\n";
        # zones entries.
        foreach my $entry ("internal", "ipsec", "dmz", "dialin")
        {
          $result .= "    <$entry enabled=\"" . ($data{config}{zones}{$entry} ? "true" : "false") . "\"\/>\n";
        }
        $result .= "    <log prefix=\"$data{config}{logInfo}{prefix}\" level=\"$data{config}{logInfo}{level}\"\/>\n";
        $result .= "    <logRate";
        foreach my $entry ("taint", "reserved", "policy", "reject", "normal")
        {
          $result .= " $entry=\"$data{config}{limitRates}{log}{$entry}\"";
        }
        $result .= "\/>\n";
        $result .= "    <limitRate icmp=\"$data{config}{limitRates}{icmp}\"\/>\n";
        $result .= "  </config>\n";

        # output <networks>
        $result .= "  <networks>\n";
        foreach my $index (sort { $a <=> $b } keys %{$data{config}{networks}})
        {
          $result .= "    <addr value=\"$data{config}{networks}{$index}{host}\" type=\"$data{config}{networks}{$index}{type}\" index=\"$index\" active=\"" . ($data{config}{networks}{$index}{active} ? "true" : "false") . "\" comment=\"$data{config}{networks}{$index}{comment}\"\/>\n";
        }
        $result .= "  </networks>\n";

        # output <reserved>
        $result .= "  <reserved>\n";
        foreach my $reserved (@{$data{config}{reserved}})
        {
          $result .= "    <addr value=\"$reserved->{host}\" active=\"" . ($reserved->{active} ? "true" : "false") . "\" limitTo=\"" . join(",", @{$reserved->{limitTo}}) . "\" comment=\"$reserved->{comment}\"\/>\n";
        }
        $result .= "  </reserved>\n";

        # output <zones>
        $result .= "  <zones>\n";
        foreach my $zone ("external", "internal", "ipsec", "dmz", "dialin")
        {
          if (scalar keys %{$data{config}{interfaces}{$zone}} == 0)
          {
            next if ($zone !~ /^(external)$/ && $data{config}{zones}{$zone} == 0);  # not being used.
          }
          else
          {
            # display the contents of the interfaces for this zone.
            $result .=  "    <zone type=\"$zone\">\n";
            foreach my $alias (sort keys %{$data{config}{interfaces}{$zone}})
            {
              my $info = $data{config}{interfaces}{$zone}{$alias};
              $result .= "      <interface name=\"$info->{name}\"";
              $result .= " alias=\"$alias\"";
              if ($zone =~ /^(external|internal|dmz|dialin)$/)
              {
                $result .=  " ip=\"$info->{ip}\" ipsec=\"" . ($info->{ipsec} ? "true" : "false") . "\"";
                $result .=  ($info->{ipsec} ? " ipsecLimitFrom=\"" . join(",", @{$info->{ipsecLimitFrom}}) . "\"" : "");
              }
              else
              {
                $result .=  " boundTo=\"$info->{boundTo}\"";
              }
              $result .=  " network=\"" . join (",", @{$info->{network}}) . "\"";
              $result .=  " proxyarp=\"" . ($info->{proxyarp} ? "true" : "false") . "\"";
              $result .=  " active=\"" . ($info->{active} ? "true" : "false") . "\" comment=\"$info->{comment}\"\/>\n";
            }
            $result .= "    <\/zone>\n";
          }
        }
        $result .= "  </zones>\n";

        # output <services>
        $result .= "  <services>\n";
        foreach my $service (@{$data{services}{servicesOrder}})
        {
          $result .=  "    <service name=\"$service\" comment=\"$data{services}{$service}{comment}\">\n";
          foreach my $info (@{$data{services}{$service}{infos}})
          {
            $result .= "      <info proto=\"$info->{proto}\"";
            $result .= " dport=\"$info->{dport}\"" if ($info->{proto} =~ /^(tcp|udp)$/);
            $result .= " sport=\"$info->{sport}\"" if ($info->{proto} =~ /^(tcp|udp)$/);
            $result .= " icmpType=\"$info->{icmpType}\"" if ($info->{proto} =~ /^(icmp)$/);
            $result .= "\/>\n";
          }
          $result .= "    </service>\n";
        }
        $result .= "  </services>\n";

        # output <paths>
        $result .= "  <paths>\n";
        foreach my $service (@{$data{paths}{serviceGroupsOrder}})
        {
          $result .=  "    <serviceGroup name=\"$service\" active=\"" . ($data{paths}{$service}{active} ? "true" : "false") . "\" comment=\"$data{paths}{$service}{comment}\">\n";
          foreach my $path (@{$data{paths}{$service}{zonePathOrder}})
          {
            foreach my $node (@{$data{paths}{$service}{$path}})
            {
              $result .=  "      <path type=\"$path\"";
              # now output the global info about this path.
              $result .= " i=\"" . join(",", @{$node->{i}}) . "\"" if (exists $node->{i});
              $result .= " o=\"" . join(",", @{$node->{o}}) . "\"" if (exists $node->{o});
              $result .= " ip=\"$node->{ip}\"" if ($node->{ip});
              $result .= " limitFrom=\"" . join(",", @{$node->{limitFrom}}) . "\"" if (exists $node->{limitFrom});
              $result .= " limitTo=\"" . join(",", @{$node->{limitTo}}) . "\"" if (exists $node->{limitTo} && $path !~ /^(firewallToFirewall)$/);
              $result .= " active=\"" . ($node->{active} ? "true" : "false") . "\" comment=\"$node->{comment}\">\n";
              $result .= "        <reject\/>\n" if ($node->{reject});
              $result .= "        <dropNoLog\/>\n" if ($node->{dropNoLog});
              $result .= "        <dnat destination=\"$node->{dnatDestination}\" zone=\"$node->{dnatZone}\" interface=\"$node->{dnatInterface}\" dnatPort=\"$node->{dnatPort}\"\/>\n" if ($node->{dnat});
              $result .= "        <redirect redirectPort=\"$node->{redirectPort}\"\/>\n" if ($node->{redirect});
              $result .= "        <queue\/>\n" if ($node->{queue});
              $result .= "        <log comment=\"$node->{logComment}\" level=\"$node->{logLevel}\"\/>\n" if ($node->{log});
              foreach my $serviceEntry (@{$node->{service}})
              {
                $result .= "        <service name=\"$serviceEntry\"\/>\n";
              }
              $result .= "      </path>\n";
            }
          }
          $result .= "    </serviceGroup>\n";
        }
        $result .= "  </paths>\n";

        $result .= "</firewall>\n";
      }
      else
      {
        die "Version = '" . $data{'_info_'}{version} . "' is invalid for module = '" . $data{'_info_'}{module} . "'!\n";
      }
    }
    else
    {
      die "module = '" . $data{'_info_'}{module} . "' is invalid!\n";
    }
  }
  else
  {
    $result .= "Config File not valid!\n\n";
    $result .= join("\n", @{$valid[1]}) . "\n";
    die $result;
  }

  return $result;
}

1;
__END__

=head1 NAME

ConfigData - The XML Configuration Data Module.

=head1 SYNOPSIS

  use PCXFireWall::ConfigData;
  my $obj = PCXFireWall::ConfigData->new();

=head1 DESCRIPTION

ConfigData will contain the parsed XML Configuration files used by PCX
Firewall Rules Modules.  It provides a method to validate that it is
complete and also a method to generate a valid XML config file from the
data stored in the data hash.

=head1 FUNCTIONS

  scalar new()
    Creates a new instance of the PCXFireWall::ConfigData
    object.

  array isValid(void)
    Determines if the data structure is complete and usable for
    generating an XML config file from.
    Returns an array.  The first index is boolean (1 or 0 to indicate
    if the object is valid or not).  The second index is an array of
    error messages that were generated based upon the errors found.

  string generateXML(void)
    Creates an XML Config file based upon the info stored in the data
    hash.  It first calls isValid() to make sure this is possible.  If
    not then we die with an informative error message.

=head1 VARIABLES

  data - hash that represents the xml config file.

  errorCodes - hash of error codes to messages returned by isValid().

  NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.
         Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:admin at pcxperience.com)

=head1 SEE ALSO

perl(1)

=cut
